import requests
import json
import tqdm
import math
import os
import zipfile as zf
import contextlib

from .exceptions import (
    InvalidGranuleException, InvalidCredentialsException, GranuleDownloadException
)
from .granules import SentinelGranule
from .pairs import SentinelGranulePair


def download(
        to_download,
        credentials,
        directory='.',
        progress_bar=False,
        unzip=True
):
    """Download sentinel granules or granule pairs from asf api

    :param to_download: granule string, SentinelGranule, or SentinelGranulePair
    :param credentials: dictionary containing the callers earthdata credentials
    :param directory: the directory where the granule will be downloaded
    :param progress_bar: show a progress_bar when downloading
    :param unzip: select if the granule should be unziped after downloading

    The credentials dictionary should look like:
    .. code-block:: python

       {'username': USERNAME, 'password': PASSWORD}

    """
    def make_dl_func_with_params(gran):
        return lambda: single_granule_download(
            gran,
            credentials,
            directory,
            progress_bar,
            unzip
        )

    if is_single_download(to_download):
        download_func = make_dl_func_with_params(to_download)
        download_func()

    elif isinstance(to_download, SentinelGranulePair):
        for g in to_download.tuple:
            download_func = make_dl_func_with_params(g)
            download_func()
    else:
        raise TypeError('Cannot download {}'.format(type(to_download)))


def is_single_download(to_download):
    try:
        # python2
        return isinstance(to_download, (SentinelGranule, str, unicode))
    except NameError:
        # python3
        return isinstance(to_download, (SentinelGranule, str))


def single_granule_download(
    granule,
    credentials,
    directory,
    progress_bar,
    unzip
):
    granule_str = str(granule)

    zip_url = get_download_url(granule_str)
    download_redirect = requests.get(zip_url)

    username, password = [credentials[k] for k in ['username', 'password']]
    dl_request = requests.get(
        download_redirect.url,
        auth=(username, password),
        stream=True
    )

    total_size = int(dl_request.headers.get('content-length', 0))
    download = get_download_stream(
        dl_request,
        progress_bar,
        total_size
    )

    if not os.path.exists(directory):
        os.makedirs(directory)

    granule_path = get_granule_path(directory, granule_str)
    do_download(download, granule_path, total_size)

    if not download_was_authorized(granule_path):
        raise InvalidCredentialsException('username or password incorrect')

    if unzip:
        do_unzip(granule_path)


def get_download_url(granule_str):
    api_url = 'https://api.daac.asf.alaska.edu/services/search/param'

    resp = requests.post(api_url, {
        'granule_list': granule_str,
        'output': 'JSON'
    })

    responses = json.loads(resp.text)[0]
    urls = [resp['downloadUrl'] for resp in responses]

    zip_urls = [url for url in urls if url.endswith('.zip')]

    return zip_urls.pop()


def get_download_stream(dl_response, progress_bar, total_size):
    block_size = 1024

    if not progress_bar:
        download = dl_response.iter_content(block_size)
    else:
        download = tqdm.tqdm(
            dl_response.iter_content(block_size),
            total=math.ceil(total_size//block_size),
            unit='B',
            unit_scale=True
        )

    return download


def get_granule_path(directory, granule_str):
    granule_zip = granule_str + '.zip'

    return os.path.join(directory, granule_zip)


def do_download(download, directory, total_size):
    with open(directory, 'wb') as f:
        wrote = stream_to_file(f, download)

    if total_size != 0 and wrote != total_size:
        raise GranuleDownloadException("ERROR downloading granule")


def stream_to_file(f, download):
    wrote = 0
    for chunk in download:
        wrote += len(chunk)
        f.write(chunk)

    return wrote


def download_was_authorized(granule_path):
    return zf.is_zipfile(granule_path)


def do_unzip(path):
    output_path = os.path.dirname(path)

    with zipfile(path) as zf:
        zf.extractall(output_path)


@contextlib.contextmanager
def zipfile(path):
    zip_ref = zf.ZipFile(path, 'r')
    yield zip_ref
    zip_ref.close()




import unittest
import os
import json

import asf_granule_util as gu


class TestDownloadPairs(unittest.TestCase):
    def setUp(self):
        path = os.path.dirname(__file__)
        test_granules_path = os.path.join(
            path, 'data/granules.json'
        )

        with open(test_granules_path, 'r') as f:
            self.ga, self.gb = json.load(f)

        self.ga_obj = gu.SentinelGranule(self.ga)
        self.gb_obj = gu.SentinelGranule(self.gb)

        self.pair = gu.SentinelGranulePair(self.ga_obj, self.gb_obj)

        self.directory = '.'

        username, password = get_creds(os.path.join(path, '..', 'creds.txt'))
        self.creds = {
            'username': username,
            'password': password
        }

    def test_pair_download(self):
        pass
        # gu.download(
            # self.pair,
            # credentials=self.creds,
            # progess_bar=True
        # )


def get_creds(path):
    with open(path, 'r') as f:
        out = f.read().strip().split()

    return out


if __name__ == "__main__":
    unittest.main()



from .exceptions import InvalidGranuleException
import re
import datetime


class Granule(object):
    """Base class granuel object. pattern needs to be a regex with the same
    number of groups as the length of attributes"""
    def __init__(self, pattern, granule_string, attributes):
        m = re.search(pattern, granule_string)
        if m is None:
            raise InvalidGranuleException(
                "The string given does not describe a valid granule!",
                granule_string
            )

        g = m.groups()
        for i, attribute in enumerate(attributes):
            try:
                setattr(self, attribute, g[i])
            except IndexError:
                break

    # @abstractmethod <- So it works with python2
    def to_str(self):
        pass

    def __str__(self):
        return str(self.to_str())

    def __repr__(self):
        return 'Granule({g})'.format(g=self)


class SentinelGranule(Granule):
    """Object for working with sentinel granules."""
    # From: https://earth.esa.int/web/sentinel/user-guides/sentinel-1-sar/naming-conventions
    pattern = re.compile(r"""
            (S1[AB])_           # Mission ID
            (IW|EW|WV|S[1-6])_  # Mode/Beam
            (GRD|SLC|OCN)       # Product Type
            ([FHM_])_           # Resolution
            ([12])              # Processing Level
            ([SA])              # Product Class
            (SH|SV|DH|DV)_      # Polarization
            (\d{8})T(\d{6})_    # Start (Date)T(Time)
            (\d{8})T(\d{6})_    # End (Date)T(Time)
            (\d{6})_            # Absolut Orbit Number
            ([0-9A-F]{6})_      # Missin Data Take ID
            ([0-9A-F]{4})       # Product Unique ID
    """, re.X)

    @staticmethod
    def is_valid(possible_granule_string):
        """Check if a string is a valid sentinel granule."""
        return re.match(SentinelGranule.pattern, possible_granule_string)

    def __init__(self, granule_string):
        length = len(granule_string)

        if length != 67:
            raise InvalidGranuleException(
                "Granule string is too {}! Must be 67 characters".format(
                    "short" if length < 67 else "long"
                ),
                granule_string
            )

        attributes = self.get_attributes()

        super(SentinelGranule, self).__init__(
            SentinelGranule.pattern,
            granule_string,
            attributes
        )

    def __eq__(self, other):
        return str(self) == str(other)

    def to_str(self):
        return "_".join([
            self.mission,
            self.beam_mode,
            self.prod_type + self.res,
            self.proc_level + self.prod_class + self.pol,
            self.start_date + "T" + self.start_time,
            self.stop_date + "T" + self.stop_time,
            self.orbit,
            self.data_id,
            self.unique_id
        ])

    def get_attributes(self):
        """Get all the attributes the granule has"""
        return [
            "mission",
            "beam_mode",
            "prod_type",
            "res",
            "proc_level",
            "prod_class",
            "pol",
            "start_date",
            "start_time",
            "stop_date",
            "stop_time",
            "orbit",
            "data_id",
            "unique_id"
        ]

    def get_start_date(self):
        """Returns datetime object start date"""
        return self.get_date(self.start_date + self.start_time)

    def get_stop_date(self):
        """Returns datetime object of stop date"""
        return self.get_date(self.stop_date + self.stop_time)

    def get_date(self, date_str):
        """Returns datetime object from date_str formatted '%Y%m%d%H%M%S'"""
        return datetime.datetime.strptime(date_str, '%Y%m%d%H%M%S')



import unittest
import os
import json

from .timeout import timeout, TimeoutException
import asf_granule_util as gu


class TestDownload(unittest.TestCase):
    def setUp(self):
        path = os.path.dirname(__file__)
        test_granules_path = os.path.join(
            path, 'data/granules.json'
        )

        with open(test_granules_path, 'r') as f:
            self.ga, self.gb = json.load(f)

        self.ga_obj = gu.SentinelGranule(self.ga)
        self.gb_obj = gu.SentinelGranule(self.gb)

        self.directory = 'granules/test'

        username, password = get_creds(os.path.join(path, '..', 'creds.txt'))
        self.creds = {
            'username': username,
            'password': password
        }

    def test_invalid_credentials_raise(self):
        with self.assertRaises(gu.InvalidCredentialsException):
            gu.download(
                to_download=self.ga_obj,
                credentials={
                    'username': 'invalid',
                    'password': 'creds'
                },
                directory=self.directory
            )

    def test_wrong_cred_keys(self):
        bad_creds = [
            {'usrname': 'hello', 'password': 'world'},
            {'username': 'hello', 'passwor': 'world'}
        ]

        for cred in bad_creds:
            with self.assertRaises(KeyError):
                gu.download(
                    to_download=self.ga_obj,
                    credentials=cred,
                    directory=self.directory
                )

        self.cleanup(self.ga_obj)

    def test_download_starts_with_bar(self):
        self.download_test(
            self.ga,
            has_bar=True
        )

    def test_download_starts_without_bar(self):
        self.download_test(
            self.gb_obj,
            has_bar=False
        )

    def download_test(self, granule, has_bar):
        with self.assertRaises(TimeoutException):
            self.download_with_timeout(granule, has_bar=has_bar)

        dl_started = self.has_download_started(granule)
        self.assertTrue(dl_started)

        self.cleanup(granule)

    def download_with_timeout(self, granule, has_bar):
        with timeout(5):
            gu.download(
                to_download=granule,
                credentials=self.creds,
                directory=self.directory,
                progress_bar=has_bar,
                unzip=False
            )

    def has_download_started(self, granule):
        path = self.get_full_download_path(granule)

        return os.path.exists(path)

    def cleanup(self, granule):
        dl_path = self.get_full_download_path(granule)

        if os.path.exists(dl_path):
            os.remove(dl_path)
            os.rmdir(self.directory)

    def get_full_download_path(self, granule):
        return os.path.join(self.directory, str(granule) + '.zip')


def get_creds(path):
    with open(path, 'r') as f:
        out = f.read().strip().split()

    return out


if __name__ == "__main__":
    unittest.main()



"""
 A collection of helper classes for dealing with everything granule related.
 Hopefully this can be expanded on in the future to add support for different
 types of granules such as the legacy granules
"""

from .granules import SentinelGranule
from .pairs import GranulePair, SentinelGranulePair
from .stack import GranuleStack, SentinelGranuleStack
from .exceptions import InvalidGranuleException, InvalidCredentialsException
from .download import download


__all__ = [
    "SentinelGranule", "GranulePair",
    "SentinelGranulePair", "GranuleStack", "SentinelGranuleStack",
    "download",
    "InvalidGranuleException", "InvalidCredentialsException"
]



import unittest
import os
import json

import asf_granule_util as gu


class TestPairs(unittest.TestCase):
    def setUp(self):
        self.data_path = os.path.join(
            os.path.dirname(__file__), 'data'
        )

        test_granules_path = os.path.join(self.data_path, 'granules.json')
        with open(test_granules_path, 'r') as f:
            self.ga, self.gb = json.load(f)

        self.ga_obj = gu.SentinelGranule(self.ga)
        self.gb_obj = gu.SentinelGranule(self.gb)

        self.pair = gu.SentinelGranulePair(
            self.ga_obj,
            self.gb_obj
        )

    def test_make_pair_keys(self):
        self.assertEqual(self.gb_obj, self.pair.master)
        self.assertEqual(self.ga_obj, self.pair.slave)

    def test_str_and_repr(self):
        path = os.path.join(self.data_path, 'correct_prints.json')
        with open(path, 'r') as f:
            correct = json.load(f)['pair']

        self.assertEqual(
            correct['str'],
            str(self.pair)
        )

        self.assertEqual(
            correct['repr'],
            repr(self.pair)
        )

    def test_tuple(self):
        g_tuple = (self.gb_obj, self.ga_obj)

        self.assertEqual(g_tuple, self.pair.tuple)


if __name__ == '__main__':
    unittest.main()



import signal


class TimeoutException(Exception):
    pass


class timeout:
    def __init__(self, seconds, error_message=None):
        if error_message is None:
            error_message = 'test timed out after {}s.'.format(seconds)
        self.seconds = seconds
        self.error_message = error_message

    def handle_timeout(self, *_):
        raise TimeoutException(self.error_message)

    def __enter__(self):
        signal.signal(signal.SIGALRM, self.handle_timeout)
        signal.alarm(self.seconds)

    def __exit__(self, exc_type, exc_val, exc_tb):
        signal.alarm(0)



import os
import unittest
import json
import datetime

import asf_granule_util as gu


class TestSentinelGranule(unittest.TestCase):
    def setUp(self):
        self.data_path = os.path.join(
            os.path.dirname(__file__), 'data'
        )

        test_granules_path = os.path.join(self.data_path, 'granules.json')
        with open(test_granules_path, 'r') as f:
            self.ga, self.gb = json.load(f)

        self.ga_obj = gu.SentinelGranule(self.ga)
        self.gb_obj = gu.SentinelGranule(self.gb)

    def test_is_valid(self):
        self.assertTrue(gu.SentinelGranule.is_valid(self.ga))
        self.assertFalse(gu.SentinelGranule.is_valid('bob'))

    def test_matches_granules(self):
        to_str_value = str(self.ga_obj)

        self.assertEqual(to_str_value, self.ga)

    def test_too_short_granules(self):
        too_short_str = 'afosd=b90'

        self.invalid_granules_raise_with(too_short_str)

    def test_invald_mission(self):
        bad_mission = self.gb.replace('B', 'C')

        self.invalid_granules_raise_with(bad_mission)

    def invalid_granules_raise_with(self, test_str):
        with self.assertRaises(gu.InvalidGranuleException):
            gu.SentinelGranule(test_str)

    def test_datetime_objects(self):
        start_date = datetime.datetime(
            year=2015, month=8, day=29, hour=12, minute=37, second=51
        )
        stop_date = datetime.datetime(
            year=2015, month=8, day=29, hour=12, minute=38, second=21
        )

        self.assertEqual(
            start_date,
            self.ga_obj.get_start_date()
        )
        self.assertEqual(
            stop_date,
            self.ga_obj.get_stop_date()
        )

    def test_date_strings(self):
        start_date, stop_date = self.ga[17:25], self.ga[33:41]

        self.assertEqual(start_date, self.ga_obj.start_date)
        self.assertEqual(stop_date, self.ga_obj.stop_date)

    def test_time_strings(self):
        start_time, stop_time = self.ga[26:32], self.ga[42:48]

        self.assertEqual(start_time, self.ga_obj.start_time)
        self.assertEqual(stop_time, self.ga_obj.stop_time)

    def test_str_and_repr(self):
        path = os.path.join(self.data_path, 'correct_prints.json')
        with open(path, 'r') as f:
            correct = json.load(f)['sentinel']

        self.assertEqual(
            correct['str'],
            str(self.ga_obj)
        )

        self.assertEqual(
            correct['repr'],
            repr(self.ga_obj)
        )


if __name__ == '__main__':
    unittest.main()



from .pairs import get_pairs, SentinelGranulePair


class GranuleStack(object):
    def __init__(self, granule_list):
        self.granules = granule_list

    def add_granule(self, granule):
        self.granules.append(granule)

    @property
    def pairs(self):
        return get_pairs(self.granules, 1) + get_pairs(self.granules, 2)

    def to_str(self):
        return [pair.to_str() for pair in self.pairs]

    def __str__(self):
        pass

    def __repr__(self):
        return self.__str__()


class SentinelGranuleStack(GranuleStack):
    @property
    def pairs(self):
        return get_pairs(self.granules, 1, SentinelGranulePair) + \
            get_pairs(self.granules, 2, SentinelGranulePair)



from .granules import SentinelGranule

class GranulePair(object):
    """Object for working with pairs of granuels"""
    def __init__(self, master, slave):
        self.master = master
        self.slave = slave

    @property
    def tuple(self):
        """Property for getting a tuple with the granules in the pair"""
        return (self.master, self.slave)

    def __str__(self):
        return "({m}, {s})".format(m=self.master, s=self.slave)

    def __repr__(self):
        return "GranulePair({m}, {s})".format(
            m=self.master,
            s=self.slave
        )


class SentinelGranulePair(GranulePair):
    """Object for working with pairs of sentinel granules."""
    @staticmethod
    def get_time_delta(master, slave):
        return master.get_start_date() - slave.get_start_date()

    def __init__(self, master, slave):
        master, slave = [SentinelGranule(str(g)) for g in (master, slave)]

        delta = SentinelGranulePair.get_time_delta(master, slave)
        if delta.days < 0:
            master, slave = slave, master

        super(SentinelGranulePair, self).__init__(
            master,
            slave,
        )

    def __eq__(self, other):
        return str(self) == str(other)

    def time_delta(self):
        """Get the difference in start times as a datetime object."""
        return SentinelGranulePair.get_time_delta(self.master, self.slave)

    def get_dates(self):
        """Get the dates of the granules as a string"""
        return self.master.start_date + "_" + self.slave.start_date

    def checksum_str(self):
        """Get the checksums of the granules as a string"""
        return "{}-{}".format(self.master.unique_id, self.slave.unique_id)


def get_pairs(granule_list, gap=1, pair_type=GranulePair):
    """pairs up granule, takes a list of granules sorted by time"""
    granule_list.sort(key=lambda g: g.start_date)
    pair_list = []

    for i, granule in enumerate(granule_list):
        try:
            pair = pair_type(granule, granule_list[i + gap])
            pair_list.append(pair)
        except Exception:
            break

    return pair_list




from setuptools import setup, find_packages

import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

long_description = '''
    Utility classes for making working with sential granule
    strings easier and more readable. Documentation can be found here:
    http://asf-granule-util-docs.s3-website-us-west-2.amazonaws.com/
'''

with codecs.open("version.txt", "r+") as f:
    version = f.read()


with codecs.open('requirements.txt', 'r') as f:
    requirements = f.read().strip().split('\n')[1:]

setup(
    name='asf_granule_util',
    version=str(version),

    description='Library for handling sential granules',
    long_description=long_description,

    url='https://github.com/asfadmin/hyp3-time-series-lib',

    author='ASF Student Development Team 2017',
    author_email='eng.accts@asf.alaska.edu',

    license="'License :: OSI Approved :: MIT License'",

    classifiers=['Development Status :: 3 - Alpha',

                 # Indicate who your project is intended for
                 'Intended Audience :: Science/Research',
                 'Topic :: Scientific/Engineering :: GIS',

                 # Specify the Python versions you support here. In particular, ensure
                 # that you indicate whether you support Python 2, Python 3 or both.
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3',
                 ],
    keywords='granule asf sential util',
    packages=find_packages(),
    install_requires=requirements

)



# exceptions.py
# Rohan Weeden
# Created: August 17, 2017


class InvalidGranuleException(Exception):
    """Custom exception thrown by the module, indicates a invalid granule string."""
    def __init__(self, string, granule_string):
        super(InvalidGranuleException, self).__init__(string)
        self.granule = granule_string


class InvalidCredentialsException(Exception):
    """Earthdata credentials were invalid when downloading granule"""
    pass


class GranuleDownloadException(Exception):
    pass
